import logging
import requests

from django.contrib.auth import get_user_model, authenticate
from django.core.management.base import BaseCommand
from django.db import connection
from django.db.utils import OperationalError
from elasticsearch.exceptions import ElasticsearchException
from redis import exceptions as redis_exceptions
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.schemas import EndpointInspector

from nodeconductor.logging.elasticsearch_client import ElasticsearchClient, ElasticsearchClientError
from nodeconductor.server.celery import app as celery_app

User = get_user_model()


class Command(BaseCommand):
    help = "Check status of Waldur MasterMind configured services"

    def add_arguments(self, parser):
        parser.add_argument(
            '--check-api-endpoints-at',
            dest='base_url', default=None,
            help='Runs API endpoints check at specified base URL (i.e. http://example.com). '
                 'If this argument is not provided, check will be skipped.',
        )

    def handle(self, *args, **options):
        success_status = self.style.SUCCESS(' [OK]')
        error_status = self.style.ERROR(' [ERROR]')
        output_messages = {
            'database': ' - Database %(vendor)s connection',
            'workers': ' - Task runners (Celery workers)',
            'redis': ' - Queue and cache server (Redis) connection',
            'elasticsearch': ' - Event store (Elasticsearch) connection',
        }
        padding = len(max(output_messages.values(), key=len))
        # If services checks didn't pass, skip API endpoints check
        skip_endpoints = False

        # Rise logging level to prevent redundant log messages
        logging.disable(logging.CRITICAL)
        self.stdout.write('Checking Waldur MasterMind services...')

        # Check database connectivity
        db_vendor = connection.vendor.capitalize().replace('sql', 'SQL').replace('Sql', 'SQL')
        self.stdout.write((output_messages['database'] % {'vendor': db_vendor}).ljust(padding), ending='')
        try:
            connection.cursor()
        except OperationalError:
            skip_endpoints = True
            self.stdout.write(error_status)
        else:
            self.stdout.write(success_status)

        # Check celery and redis
        celery_inspect = celery_app.control.inspect()
        celery_results = {
            'workers': success_status,
            'redis': success_status,
        }
        try:
            stats = celery_inspect.stats()
            if not stats:
                skip_endpoints = True
                celery_results['workers'] = error_status
        except redis_exceptions.RedisError:
            skip_endpoints = True
            celery_results['redis'] = error_status
            celery_results['workers'] = error_status
        finally:
            self.stdout.write(output_messages['workers'].ljust(padding) + celery_results['workers'])
            self.stdout.write(output_messages['redis'].ljust(padding) + celery_results['redis'])

        # Check ElasticSearch
        self.stdout.write(output_messages['elasticsearch'].ljust(padding), ending='')
        try:
            es_client = ElasticsearchClient()
            if es_client.client.ping():
                self.stdout.write(success_status)
            else:
                skip_endpoints = True
                self.stdout.write(error_status)
        except (ElasticsearchException, ElasticsearchClientError):
            skip_endpoints = True
            self.stdout.write(error_status)

        if skip_endpoints:
            self.stderr.write('API endpoints check skipped due to erred services')
            exit(1)
        elif options['base_url'] is None:
            self.stdout.write('API endpoints check skipped')
        else:
            self._check_api_endpoints(options['base_url'])

        # return logging level back
        logging.disable(logging.NOTSET)

    def _check_api_endpoints(self, base_url):
        self.stdout.write('\nChecking Waldur MasterMind API endpoints...')
        inspector = EndpointInspector()
        endpoints = inspector.get_api_endpoints()
        user, _ = User.objects.get_or_create(username='waldur_status_checker', is_staff=True)
        authenticate(username='waldur_status_checker')
        token = Token.objects.get(user=user)

        for endpoint in endpoints:
            path, method, view = endpoint
            if method != 'GET' or '{pk}' in path or '{uuid}' in path:
                continue

            url = base_url + path
            self.stdout.write(' Checking %s endpoint...' % url, ending='')
            try:
                response = requests.get(url, headers={'Authorization': 'Token %s' % token.key})
            except requests.RequestException:
                self.stdout.write(self.style.ERROR(' [ERROR]'))
            else:
                if response.status_code != status.HTTP_200_OK:
                    self.stdout.write(self.style.ERROR(' [%d]' % response.status_code))
                else:
                    self.stdout.write(self.style.SUCCESS(' [200]'))

        # clean up
        user.delete()



from nodeconductor.logging.loggers import EventLogger, event_logger

from . import models


class IssueEventLogger(EventLogger):
    issue = models.Issue

    class Meta:
        event_types = ('issue_deletion_succeeded',
                       'issue_update_succeeded',
                       'issue_creation_succeeded')
        event_groups = {
            'support': event_types,
        }


class OfferingEventLogger(EventLogger):
    offering = models.Offering

    class Meta:
        event_types = (
            'offering_created',
            'offering_deleted',
            'offering_state_changed',
        )
        event_groups = {
            'support': event_types,
        }


event_logger.register('waldur_issue', IssueEventLogger)
event_logger.register('waldur_offering', OfferingEventLogger)



import django_filters

from nodeconductor.core import filters as core_filters

from . import models


class SlurmPackageFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr='icontains')
    service_settings = core_filters.URLFilter(
        view_name='servicesettings-detail', name='service_settings__uuid')

    class Meta(object):
        model = models.SlurmPackage
        fields = []



from __future__ import unicode_literals

from nodeconductor_assembly_waldur.billing import views


def register_in(router):
    router.register(r'billing-price-estimates', views.PriceEstimateViewSet, base_name='billing-price-estimate')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-07-11 07:41
from __future__ import unicode_literals

from decimal import Decimal
from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import nodeconductor.core.fields
import nodeconductor.structure.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('structure', '0052_customer_subnets'),
        ('experts', '0002_expertrequest'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExpertBid',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', nodeconductor.core.fields.UUIDField()),
                ('price', models.DecimalField(decimal_places=7, default=0, max_digits=22, validators=[django.core.validators.MinValueValidator(Decimal('0'))])),
                ('request', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='bids', to='experts.ExpertRequest')),
                ('team', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Project')),
                ('user', models.ForeignKey(help_text='The user which has created this bid.', on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-created'],
            },
            bases=(nodeconductor.structure.models.StructureLoggableMixin, models.Model),
        ),
    ]



#!/usr/bin/env python
from setuptools import setup, find_packages

# defusedxml is required by djangosaml2
install_requires = [
    'ansible-waldur-module>=0.4.1',
    'defusedxml>=0.4.1',
    'influxdb>=4.1.0',
    'jira>=1.0.7',
    'nodeconductor>=0.148.3',
    'nodeconductor_auth_social>=0.7.3',
    'nodeconductor_auth_openid>=0.8.4',
    'nodeconductor_aws>=0.10.0',
    'nodeconductor_azure>=0.3.0',
    'nodeconductor_cost_planning>=0.5.0',
    'nodeconductor_digitalocean>=0.9.0',
    'nodeconductor_openstack>=0.37.3',
    'nodeconductor_saml2>=0.8.1',
    'waldur_ansible>=0.3.0',
    'waldur_freeipa>=0.2.2',
    'waldur_paypal>=0.6.2',
    'waldur_slurm>=0.2.7',
]

test_requires = [
    'ddt>=1.0.0,<1.1.0',
    'factory_boy==2.4.1',
    'freezegun==0.3.7',
]

setup(
    name='nodeconductor-assembly-waldur',
    version='2.7.8',
    author='OpenNode Team',
    author_email='info@opennodecloud.com',
    url='http://waldur.com',
    description='Waldur MasterMind is a hybrid cloud orchestrator.',
    license='MIT',
    long_description=open('README.rst').read(),
    package_dir={'': 'src'},
    packages=find_packages('src', exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),
    install_requires=install_requires,
    extras_require={
        'test': test_requires,
    },
    zip_safe=False,
    entry_points={
        'nodeconductor_extensions': (
            'waldur_packages = nodeconductor_assembly_waldur.packages.extension:PackagesExtension',
            'waldur_invoices = nodeconductor_assembly_waldur.invoices.extension:InvoicesExtension',
            'waldur_support = nodeconductor_assembly_waldur.support.extension:SupportExtension',
            'waldur_analytics = nodeconductor_assembly_waldur.analytics.extension:AnalyticsExtension',
            'waldur_experts = nodeconductor_assembly_waldur.experts.extension:ExpertsExtension',
            'waldur_billing = nodeconductor_assembly_waldur.billing.extension:BillingExtension',
            'waldur_slurm_invoices = nodeconductor_assembly_waldur.slurm_invoices.extension:SlurmInvoicesExtension',
        ),
    },
    include_package_data=True,
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
)



from django.db import models as django_models

from nodeconductor.core import managers as core_managers


class GenericInvoiceItemManager(core_managers.GenericKeyMixin, django_models.Manager):
    pass



from nodeconductor_openstack.openstack.models import Tenant
from nodeconductor_openstack.openstack_tenant.models import Instance


def get_tenants():
    if not Tenant.objects.exists():
        return []
    points = []
    tenants = Tenant.objects.all()
    quota_names = Tenant.get_quotas_names()
    quotas = Tenant.get_sum_of_quotas_as_dict(tenants, quota_names=quota_names)

    for quota_name in quota_names:
        points.append({
            'measurement': 'openstack_%s' % quota_name,
            'fields': {
                'limit': quotas[quota_name],
                'usage': quotas['%s_usage' % quota_name],
            }
        })
    return points


def get_instances():
    states = count_instances_by_state(Instance.objects.all())

    points = []
    for state, count in states.items():
        points.append({
            'measurement': 'openstack_instance_runtime_state',
            'tags': {
                'state': state,
            },
            'fields': {
                'count': count,
            }
        })
    return points


def count_instances_by_state(instances):
    erred = 0
    online = 0
    offline = 0
    provisioning = 0

    for instance in instances.values('state', 'runtime_state'):
        if instance['state'] == Instance.States.ERRED:
            erred += 1
        elif instance['state'] != Instance.States.OK:
            provisioning += 1
        elif instance['runtime_state'] == Instance.get_online_state():
            online += 1
        elif instance['runtime_state'] == Instance.get_offline_state():
            offline += 1

    return {
        'erred': erred,
        'online': online,
        'offline': offline,
        'provisioning': provisioning,
        'total': erred + online + offline + provisioning
    }



from __future__ import unicode_literals

from nodeconductor.core import NodeConductorExtension


class BillingExtension(NodeConductorExtension):
    @staticmethod
    def django_app():
        return 'nodeconductor_assembly_waldur.billing'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def is_assembly():
        return True



import django_filters

from nodeconductor.core import filters as core_filters

from . import models


class ExpertProviderFilter(django_filters.FilterSet):
    customer = core_filters.URLFilter(view_name='customer-detail', name='customer__uuid')
    customer_uuid = django_filters.UUIDFilter(name='customer__uuid')

    class Meta(object):
        model = models.ExpertProvider
        fields = []


class ExpertRequestFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr='icontains')
    project = core_filters.URLFilter(view_name='project-detail', name='project__uuid')
    project_uuid = django_filters.UUIDFilter(name='project__uuid')
    o = django_filters.OrderingFilter(fields=(
        ('name', 'name'),
        ('type', 'type'),
        ('state', 'state'),
        ('project__customer__name', 'customer_name'),
        ('project__name', 'project_name'),
        ('created', 'created'),
        ('modified', 'modified'),
    ))

    class Meta(object):
        model = models.ExpertRequest
        fields = ['state']


class ExpertBidFilter(django_filters.FilterSet):
    request = core_filters.URLFilter(view_name='expert-request-detail', name='request__uuid')
    request_uuid = django_filters.UUIDFilter(name='request__uuid')
    customer = core_filters.URLFilter(view_name='customer-detail', name='team__customer__uuid')
    customer_uuid = django_filters.UUIDFilter(name='team__customer__uuid')

    class Meta(object):
        model = models.ExpertBid
        fields = []



from __future__ import unicode_literals

from datetime import datetime
import functools
import re
from six.moves.html_parser import HTMLParser
import json

from django.conf import settings
from django.utils import six
from jira import JIRA, JIRAError, Comment
from jira.utils import json_loads

from nodeconductor_assembly_waldur.support import models
from nodeconductor_assembly_waldur.support.backend import SupportBackendError, SupportBackend
from nodeconductor_assembly_waldur.support.log import event_logger


class JiraBackendError(SupportBackendError):
    pass


def reraise_exceptions(func):
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except JIRAError as e:
            six.reraise(JiraBackendError, e)

    return wrapped


class JiraBackend(SupportBackend):
    credentials = settings.WALDUR_SUPPORT.get('CREDENTIALS', {})
    project_settings = settings.WALDUR_SUPPORT.get('PROJECT', {})
    issue_settings = settings.WALDUR_SUPPORT.get('ISSUE', {})

    @property
    @reraise_exceptions
    def manager(self):
        # manager will be the same for all issues - we can cache it on the class level.
        if not hasattr(self.__class__, '_manager'):
            self.__class__._manager = JIRA(
                server=self.credentials['server'],
                options={'verify': self.credentials['verify_ssl']},
                basic_auth=(self.credentials['username'], self.credentials['password']),
                validate=False)
        return self.__class__._manager

    @reraise_exceptions
    def _get_field_id_by_name(self, field_name):
        if not hasattr(self.__class__, '_fields'):
            self.__class__._fields = self.manager.fields()
        try:
            return next(f['id'] for f in self.__class__._fields if field_name in f['clauseNames'])
        except StopIteration:
            return JiraBackendError('Field "{0}" does not exist in JIRA.'.format(field_name))

    def _issue_to_dict(self, issue):
        """ Convert issue to dict that can be accepted by JIRA as input parameters """
        caller = issue.caller.full_name or issue.caller.username
        parser = HTMLParser()
        args = {
            'project': self.project_settings['key'],
            'summary': parser.unescape(issue.summary),
            'description': parser.unescape(issue.description),
            'issuetype': {'name': issue.type},
            self._get_field_id_by_name(self.issue_settings['caller_field']): caller,
        }

        if issue.reporter:
            args[self._get_field_id_by_name(self.issue_settings['reporter_field'])] = issue.reporter.name
        if issue.impact:
            args[self._get_field_id_by_name(self.issue_settings['impact_field'])] = issue.impact
        if issue.priority:
            args['priority'] = {'name': issue.priority}
        return args

    @reraise_exceptions
    def create_issue(self, issue):
        backend_issue = self.manager.create_issue(**self._issue_to_dict(issue))
        if issue.assignee:
            self.manager.assign_issue(backend_issue.key, issue.assignee.backend_id)
        issue.key = backend_issue.key
        issue.backend_id = backend_issue.key
        issue.resolution = backend_issue.fields.resolution or ''
        issue.status = backend_issue.fields.status.name or ''
        issue.link = backend_issue.permalink()
        issue.priority = backend_issue.fields.priority.name
        issue.first_response_sla = self._get_first_sla_field(backend_issue)
        issue.save()

        event_logger.waldur_issue.info(
            'Issue {issue_key} has been created.',
            event_type='issue_creation_succeeded',
            event_context={
                'issue': issue,
            })

        return backend_issue

    def _get_first_sla_field(self, backend_issue):
        field_name = self._get_field_id_by_name(self.issue_settings['sla_field'])
        value = getattr(backend_issue.fields, field_name, None)
        if value and hasattr(value, 'ongoingCycle'):
            epoch_milliseconds = value.ongoingCycle.breachTime.epochMillis
            if epoch_milliseconds:
                return datetime.fromtimestamp(epoch_milliseconds / 1000.0)

    @reraise_exceptions
    def update_issue(self, issue):
        backend_issue = self.manager.issue(issue.backend_id)
        backend_issue.update(summary=issue.summary, description=issue.description)

    @reraise_exceptions
    def delete_issue(self, issue):
        backend_issue = self.manager.issue(issue.backend_id)
        backend_issue.delete()

    def _prepare_comment_message(self, comment):
        """
        Prepends user info to the comment description to display comment author in JIRA.
        User info format - '[user.full_name user.civil_number]: '.
        """
        return '[%s %s]: %s' % (comment.author.user.full_name or comment.author.user.username,
                                comment.author.user.civil_number or '',
                                comment.description)

    def extract_comment_message(self, comment_body):
        """
        Extracts comment message from JIRA comment which contains user's info in its body.
        """
        match = re.search('^(\[.*?\]\:\s)', comment_body)
        return comment_body.replace(match.group(0), '') if match else comment_body

    @reraise_exceptions
    def create_comment(self, comment):
        backend_comment = self.manager.add_comment(comment.issue.backend_id, self._prepare_comment_message(comment))
        comment.backend_id = backend_comment.id
        comment.save(update_fields=['backend_id'])

    @reraise_exceptions
    def update_comment(self, comment):
        backend_comment = self.manager.comment(comment.issue.backend_id, comment.backend_id)
        backend_comment.update(body=self._prepare_comment_message(comment))

    @reraise_exceptions
    def delete_comment(self, comment):
        backend_comment = self.manager.comment(comment.issue.backend_id, comment.backend_id)
        backend_comment.delete()

    @reraise_exceptions
    def get_users(self):
        users = self.manager.search_assignable_users_for_projects('', self.project_settings['key'], maxResults=False)
        return [models.SupportUser(name=user.displayName, backend_id=user.key) for user in users]


class ServiceDeskBackend(JiraBackend):
    servicedeskapi_path = 'servicedeskapi'

    @reraise_exceptions
    def create_comment(self, comment):
        backend_comment = self._add_comment(
            comment.issue.backend_id,
            self._prepare_comment_message(comment),
            is_internal=not comment.is_public,
        )
        comment.backend_id = backend_comment.id
        comment.save(update_fields=['backend_id'])

    def _add_comment(self, issue, body, is_internal):
        data = {
            'body': body,
            'properties': [{'key': 'sd.public.comment', 'value': {'internal': is_internal}}, ]
        }

        url = self.manager._get_url('issue/{0}/comment'.format(issue))
        response = self.manager._session.post(url, data=json.dumps(data))

        comment = Comment(self.manager._options, self.manager._session, raw=json_loads(response))
        return comment

    def expand_comments(self, issue_key):
        """
        Returns a list of comments expanded with the 'properties' attribute.
        An attribute is taken from JIRA by adding an 'expand' query parameter to the issue GET URL.
        A method is required as Jira does not indicate internal status of comments.
        More info: https://jira.atlassian.com/browse/JSD-1261.
        :param issue_key: issue key to get comments from.
        :return: list of Jira comments as dictionaries.
        """
        url = self.manager._get_url('issue/%s/comment?expand=properties' % issue_key)
        response = self.manager._session.get(url)

        backend_comments = json.loads(response.text)['comments']
        return backend_comments

    @reraise_exceptions
    def create_issue(self, issue):
        if not issue.caller.email:
            return

        # customer will be associated with the issue by updating issue arguments in _issue_to_dict.
        self._create_customer(issue.caller.email, issue.caller.full_name)
        super(ServiceDeskBackend, self).create_issue(issue)

    def _issue_to_dict(self, issue):
        args = super(ServiceDeskBackend, self)._issue_to_dict(issue)
        args[self._get_field_id_by_name(self.issue_settings['caller_field'])] = [{
            "name": issue.caller.email,
            "key": issue.caller.email
        }]
        return args

    def _create_customer(self, email, full_name):
        """
        Creates customer in Jira Service Desk without assigning it to any particular service desk.
        :param email: customer email
        :param full_name: customer full name
        :return: True if customer is created. False if user exists already.
        """
        data = {
            "fullName": full_name,
            "email": email
        }

        headers = {
            'X-ExperimentalApi': 'true',
        }

        url = "{host}rest/{path}/customer".format(host=self.credentials['server'], path=self.servicedeskapi_path)
        try:
            self.manager._session.post(url, data=json.dumps(data), headers=headers)
        except JIRAError as e:
            # TODO [TM:1/11/17] replace it with api call when such an ability is provided
            if e.status_code == 400 and "already exists" in e.text:
                return False
            else:
                raise e
        else:
            return True
